# ==================================================================
# Gossamer Threads Module Library - http://gossamer-threads.com/
#
#   GT::CGI::EventLoop
#   Author  : Scott Beck
#   $Id: EventLoop.pm,v 1.1 2001/12/11 02:50:32 sbeck Exp $
#
# Copyright (c) 2000 Gossamer Threads Inc.  All Rights Reserved.
# ==================================================================
#
# Description: Impliments an EventLoop API for CGI programming
#

package GT::CGI::EventLoop;
# ==================================================================

use vars qw/$ATTRIBS @EXPORT_OK %EXPORT_TAGS/;
use strict;
use bases 'GT::Base' => '';
use constants
    STOP  => 1,
    EXIT  => 2,
    CONT  => 3,
    HEAP  => 0,
    EVENT => 1,
    IN    => 2,
    CGI   => 3,
    ACC   => 4,
    ARG0  => 5,
    ARG1  => 6,
    ARG2  => 7,
    ARG3  => 8,
    ARG4  => 9,
    ARG5  => 10,
    ARG6  => 11,
    ARG7  => 12,
    ARG8  => 13,
    ARG9  => 14;

use Exporter;
*import = *Exporter::import;

$ATTRIBS = {
    do                        => 'do',
    default_page              => 'home',
    default_group             => undef,
    init_events               => undef,
    default_page_pre_events   => undef,
    default_page_post_events  => undef,
    default_group_pre_events  => undef,
    default_group_post_events => undef,
    needs_array_input         => undef,
    plugin_object             => undef,
    templates_path            => undef,
    pre_package               => '',
    cgi                       => undef,
    in                        => {},
    heap                      => {},
    page_pre_events           => {},
    page_post_events          => {},
    group_pre_events          => {},
    group_post_events         => {},
    groups                    => {},
    group                     => undef,
    page                      => undef,
    print_page                => \&GT::CGI::EventLoop::print_page,
    cookies                   => []
};

@EXPORT_OK = qw/
    HEAP EVENT IN CGI
    ARG0 ARG1 ARG2 ARG3 ARG4 ARG5 ARG6 ARG7 ARG8 ARG9
/;

%EXPORT_TAGS = (
    all     => [@EXPORT_OK],
    returns => [qw/STOP EXIT CONT/],
    args    => [qw/
        HEAP EVENT IN CGI
        ARG0 ARG1 ARG2 ARG3 ARG4 ARG5 ARG6 ARG7 ARG8 ARG9
    /]
);

sub init {
# --------------------------------------------------------------------
    my $self = shift;
    $self->set( @_ ) if @_;

    for ( qw/
        default_page_pre_events
        default_page_post_events
        default_group_pre_events
        default_group_post_events
        init_events
    /)
    {
        if ( defined $self->{$_} ) {
            if ( ref( $self->{$_} ) ne 'ARRAY' ) {
                $self->{$_} = [ $self->{$_} ];
            }
        }
    }
    if ( $self->{cgi} ) {
        for ( $self->{cgi}->param ) {
            my @val = $self->{cgi}->param($_);
            my $val;
            my $match;
            for my $field ( @{$self->{needs_array_input}} ) {
                if ( $_ eq $field ) {
                    $match = 1;
                    last;
                }
            }
            if ( !$match ) {
                $val = $val[0];
            }
            else {
                $val = \@val;
            }
            $self->{in}{$_} = $val;
        }
    }
}

sub mainloop {
# --------------------------------------------------------------------
    my $self = shift;
    $self->init( @_ ) if @_;

    if ( $self->{all} ) {
        my $ret = $self->yield( $self->{all} );
        if ( $ret == EXIT ) { return }
    }
    $self->{group} ||= $self->{in}{$self->{do}};
    my $orig_group = $self->{group};
    my $match = 1;
    LOOP: {{ # FIXME Add infinite recursion checks!
        for ( keys %{$self->{groups}} ) {
            if ( index( $_, $self->{group} ) == 0 ) {
                if ( exists $self->{group_pre_events}{$_} ) {
                    $self->yield( $self->{group_pre_events}{$_} );
                    return if $self->{status} == EXIT;
                
                    if ( $self->{group} ne $orig_group ) {
                        redo LOOP;
                    }
                }
                $self->yield( $self->{groups}{$_} );
                if ( $self->{group} ne $orig_group ) {
                    redo LOOP;
                }
                if ( exists $self->{groups_post_events}{$_} ) {
                    $self->yield( $self->{group_post_events}{$_} );
                    return if $self->{status} == EXIT;
                    if ( $self->{group} ne $orig_group ) {
                        redo LOOP;
                    }
                }
                last LOOP;
            }
        }

# Default group
        $self->yield( $self->{default_group} );
        if ( $self->{group} ne $orig_group ) {
            redo LOOP;
        }
    }}
    $self->{page} ||= $self->{default_page};
    my $orig_page = $self->{page};
    LOOP: {{
        if ( exists $self->{page_pre_events}{$self->{page}} ) {
            $self->yield( $self->{page_pre_events}{$self->{page}} );
            return if $self->{status} == EXIT;
            if ( $self->{page} ne $orig_page ) {
                redo LOOP;
            }
        }
    }}
    $self->{print_page}->( $self );

# Run post page events, can't change the page on a post event
    if ( exists $self->{page_post_events}{$self->{page}} ) {
        $self->yield( $self->{page_post_events}{$self->{page}} );
    }
}

sub stop { $_[0]->{status} = STOP }
sub exit { $_[0]->{status} = EXIT }
sub cont { $_[0]->{status} = CONT }

sub cookie_jar {
# --------------------------------------------------------------------
# $obj->cookie_jar($cookie_object);
# ---------------------------------
#   Stores cookies for printing when print_page is called.
#   $cookie_object should be a GT::CGI::Cookie object. Passing undef
#   will empty the cookies array ref.
#
    my $self = shift;
    if ( !defined( $_[0] ) and @_ > 0 ) {
        $self->{cookies} = [];
    }
    elsif ( @_ > 0 ) {
        push( @{$self->{cookies}}, $_[0] );
    }
    return $self->{cookies};
}

sub add_hidden {
# --------------------------------------------------------------------
    my $self = shift;
    if ( @_ and !defined( $_[0] ) ) {
        $self->{hidden} = {};
    }
    elsif ( @_ ) {
        $self->{hidden}{$_[0]} = $_[1] if @_ > 0;
    }
}

sub get_url_hidden {
# --------------------------------------------------------------------
    my ( $self ) = @_;
    my $ret = '';
    for ( keys %{$self->{hidden}} ) {
        next unless defined $self->{hidden}{$_};
        $ret .= $self->{cgi}->escape( $_ ).'='.$self->{cgi}->escape( $self->{hidden}{$_} ).';';
    }
    return $ret;
}

sub get_form_hidden {
# --------------------------------------------------------------------
    my ( $self ) = @_;
    my $ret = '';
    for ( keys %{$self->{hidden}} ) {
        next unless defined $self->{hidden}{$_};
        $ret .= '<input type="hidden" name="'.$self->{cgi}->html_escape( $_ ).'" value="'.$self->{cgi}->html_escape( $self->{hidden}{$_} ).'">';
    }
    return $ret;
}

sub page {
# --------------------------------------------------------------------
    my $self = shift;
    if ( @_ > 0 ) {
        if ( defined( $_[0] ) and -e "$self->{template_path}/$_[0].htm" ) {
            $self->{page} = "$_[0].htm";
        }
        elsif ( defined( $_[0] ) and -e "$self->{template_path}/$_[0].html" ) {
            $self->{page} = "$_[0].html";
        }
        else {
            $self->{page} = $_[0];
        }
    }
    return $self->{page};
}

sub tags {
# --------------------------------------------------------------------
    my ( $self, %tags ) = @_;
    for ( keys %tags ) {
        $self->{tags}{$_} = $tags{$_};
    }
    return $self->{tags};
}

sub default_tags {
# --------------------------------------------------------------------
    my ( $self, %tags ) = @_;

    my $set;
    for ( keys %tags ) {
        $set->{$_} = ( defined( $self->{in}{$_} ) and length( $self->{in}{$_} ) ? $self->{in}{$_} : $tags{$_} );
    }
    $self->tags( $set );
}

sub print_page {
# --------------------------------------------------------------------
    my ( $self ) = @_;
    my $form_hidden = $self->get_form_hidden;
    my $url_hidden  = $self->get_url_hidden;
    my $tags        = $self->tags( { url_hidden => \$url_hidden, form_hidden => \$form_hidden } );
    my $page        = $self->page || 'index.htm';

# Cookies can be set with CGI input
    my $cookies = [];
    if ( $self->{in}{'set-cookie'} ) {
        foreach my $key ( keys %{$self->{in}} ) {
            if ( $key =~ /^cookie-(.*)/ ) {
                push @$cookies, $self->{cgi}->cookie( -name => $1, -value => $self->{in}{$key}, -path => '/' );
            }
        }
    }

# See if we have any cookies in out cookie jar (used through program operation to set cookies without printing
# a header)
    if ( @{$self->cookie_jar} ) {
        push @$cookies, @{$self->cookie_jar};
    }

# If we have cookie header to print print them
    print @{$cookies} ? $self->{cgi}->header( -cookie => $cookies ) : $self->{cgi}->header();

    my $base = $self->{template_path};

# Make sure the template exists and is readable
    -e "$base/$page" or die "No page ($base/$page)";
    -r _ or die "Page isn't readable by this process ($< $>) ($base/$page)";

    require GT::Template;
    GT::Template->parse( $page, $tags, {
        root      => $base,
        escape    => 1,
        print     => 1,
        heap      => [ $self->func_args ]
    } );

}

sub page_pre_events {
# --------------------------------------------------------------------
    my ( $self, %in ) = @_;
    if ( keys %in ) {
        $self->{page_pre_events} = {};
        for ( keys %in ) {
            my $val = ref( $in{$_} ) eq 'ARRAY' ? $in{$_} : [ $in{$_} ];
            $self->{page_pre_events}{$_} = $val;
        }
    }
    return $self->{page_pre_events};
}

sub page_post_events {
# --------------------------------------------------------------------
    my ( $self, %in ) = @_;
    if ( keys %in ) {
        $self->{page_post_events} = {};
        for ( keys %in ) {
            my $val = ref( $in{$_} ) eq 'ARRAY' ? $in{$_} : [ $in{$_} ];
            $self->{page_post_events}{$_} = $val;
        }
    }
    return $self->{page_post_events};
}

sub group_pre_events {
# --------------------------------------------------------------------
    my ( $self, %in ) = @_;
    if ( keys %in ) {
        $self->{group_pre_events} = {};
        for ( keys %in ) {
            my $val = ref( $in{$_} ) eq 'ARRAY' ? $in{$_} : [ $in{$_} ];
            $self->{group_pre_events}{$_} = $val;
        }
    }
    return $self->{group_pre_events};
}

sub group_post_events {
# --------------------------------------------------------------------
    my ( $self, %in ) = @_;
    if ( keys %in ) {
        $self->{group_post_events} = {};
        for ( keys %in ) {
            my $val = ref( $in{$_} ) eq 'ARRAY' ? $in{$_} : [ $in{$_} ];
            $self->{group_post_events}{$_} = $val;
        }
    }
    return $self->{group_post_events};
}

sub yield {
# --------------------------------------------------------------------
    my ( $self, $pfunc, @args ) = @_;
    my $val = ref( $pfunc ) eq 'ARRAY' ? $pfunc : [ $pfunc ];
    for ( @$pfunc ) {
        my $ret;
        if ( !ref( $_ ) ) {
            my ( $pkg, $func );
            if ( index( $pfunc, '::' ) != -1 ) {
                ($pkg, $func) = $pfunc =~ /^(.*)::(.*)$/;
            }
            else {
                $func = $pfunc;
            }
            defined( $func ) or die "No function in $pfunc\n";
            $pkg = $self->{pre_package}.$pkg if $self->{pre_package};
            unless ( defined %{$pkg . '::'} ) {
                eval "require $pkg";
                die $@ if $@;
            }
            $ret = &{$pkg.'::'.$func}( $self->func_args(@args) );
        }
        elsif ( ref( $_ ) eq 'CODE' ) {
            $ret = $_->( $self->func_args(@args) );
        }
        if ( $ret == STOP or $ret == EXIT ) {
            return $ret;
        }
    }
    return CONT;
}

sub func_args { $_[0]->{heap}, $_[0], $_[0]->{in}, $_[0]->{cgi}, @_[1 .. $#_] }

1;


